# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.db.models import F
from django.db.models import Q

# Local application / specific library imports
from machina.core.db.models import get_model
from machina.core.loading import get_class

Forum = get_model('forum', 'Forum')
ForumReadTrack = get_model('forum_tracking', 'ForumReadTrack')
TopicReadTrack = get_model('forum_tracking', 'TopicReadTrack')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class TrackingHandler(object):
    """
    The TrackingHandler allows to filter list of forums and list of topics
    in order to get only the forums which contain unread topics or the unread
    topics.
    """
    def __init__(self, request=None):
        self.request = request
        self.perm_handler = request.forum_permission_handler if request \
            else PermissionHandler()

    def get_unread_forums(self, forums, user):
        """
        Returns a list of unread forums for the given user from a given
        set of forums.
        """
        unread_forums = []
        processed_forums = []

        # A user which is not authenticated will never see a forum as unread
        if not user.is_authenticated():
            return unread_forums

        # Forums are sorted by their position in the forums tree in order
        # to limit the number of operations required to determine the unread
        # forums below a forum which has the smallest position among the
        # forums from the initial set.
        # In fact, a forum is unread if at least one of its descendants
        # (including itself) is unread.
        # So the forums are sorted and then are processed by batches: from
        # a forum with a small level and all of its descendants, we get the
        # unread forums, and so on.
        sorted_forums = sorted(forums, key=lambda forum: forum.level)

        for forum in sorted_forums:
            if forum not in unread_forums and forum not in processed_forums:
                readable_forums = self.perm_handler.forum_list_filter(
                    forum.get_descendants(include_self=True), user)
                unread = ForumReadTrack.objects.get_unread_forums_from_list(readable_forums, user)
                unread_forums.extend(unread)
                processed_forums.extend(readable_forums)

        return unread_forums

    def get_unread_topics(self, topics, user):
        """
        Returns a list of unread topics for the given user from a given
        set of topics.
        """
        unread_topics = []

        # A user which is not authenticated will never see a topic as unread.
        # If there are no topics to consider, we stop here.
        if not user.is_authenticated() or topics is None or not len(topics):
            return unread_topics

        # Note: this is done to Prevent MySQL NotSupportedError errors with
        # LIMIT & IN/ALL/ANY/SOME subqueries.
        topic_ids = list(topics.values_list('id', flat=True))

        # A topic can be unread if a track for itself exists with a mark time that
        # is less important than its update date.
        topic_tracks = TopicReadTrack.objects.filter(topic__in=topic_ids, user=user)
        tracked_topics = topic_tracks.values_list('topic__pk', flat=True)

        if topic_tracks.exists():
            tracks_dict = dict(topic_tracks.values_list('topic__pk', 'mark_time'))
            for topic in topics:
                topic_last_modification_date = topic.last_post_on or topic.created
                if topic.id in tracks_dict.keys() and topic_last_modification_date > tracks_dict[topic.id]:
                    unread_topics.append(topic)

        # A topic can be unread if a track for its associated forum exists with
        # a mark time that is less important than its creation or update date.
        forum_ids = [topic.forum_id for topic in topics]
        forum_tracks = ForumReadTrack.objects.filter(forum_id__in=forum_ids, user=user)
        tracked_forums = forum_tracks.values_list('forum__pk', flat=True)

        if forum_tracks.exists():
            tracks_dict = dict(forum_tracks.values_list('forum__pk', 'mark_time'))
            for topic in topics:
                topic_last_modification_date = topic.last_post_on or topic.created
                if ((topic.forum_id in tracks_dict.keys() and topic.id not in tracked_topics) and
                        topic_last_modification_date > tracks_dict[topic.forum_id]):
                    unread_topics.append(topic)

        # A topic can be unread if no tracks exists for it
        for topic in topics:
            if topic.forum_id not in tracked_forums and topic.id not in tracked_topics:
                unread_topics.append(topic)

        return list(set(unread_topics))

    def mark_forums_read(self, forums, user):
        """
        Marks a list of forums as read.
        """
        if not forums or not user.is_authenticated():
            return

        forums = sorted(forums, key=lambda f: f.level)

        # Update all forum tracks to the current date for the considered forums
        for forum in forums:
            forum_track = ForumReadTrack.objects.get_or_create(forum=forum, user=user)[0]
            forum_track.save()
        # Delete all the unnecessary topic tracks
        TopicReadTrack.objects.filter(topic__forum__in=forums, user=user).delete()
        # Update parent forum tracks
        self._update_parent_forum_tracks(forums[0], user)

    def mark_topic_read(self, topic, user):
        """
        Marks a topic as read.
        """
        if not user.is_authenticated():
            return

        forum = topic.forum
        try:
            forum_track = ForumReadTrack.objects.get(forum=forum, user=user)
        except ForumReadTrack.DoesNotExist:
            forum_track = None

        if forum_track is None or (topic.last_post_on and forum_track.mark_time < topic.last_post_on):
            topic_track, created = TopicReadTrack.objects.get_or_create(topic=topic, user=user)
            if not created:
                topic_track.save()  # mark_time filled

            # If no other topic is unread inside the considered forum, the latter should also
            # be marked as read.
            unread_topics = forum.topics.filter(
                Q(tracks__user=user, tracks__mark_time__lt=F('last_post_on')) |
                Q(forum__tracks__user=user, forum__tracks__mark_time__lt=F('last_post_on'), tracks__isnull=True)).exclude(id=topic.id)

            if not unread_topics.exists():
                # The topics that are marked as read inside the forum for the given user
                # wil be deleted while the forum track associated with the user must be
                # created or updated.
                TopicReadTrack.objects.filter(topic__forum=forum, user=user).delete()
                forum_track, _ = ForumReadTrack.objects.get_or_create(forum=forum, user=user)
                forum_track.save()

                # Update parent forum tracks
                self._update_parent_forum_tracks(forum, user)

    def _update_parent_forum_tracks(self, forum, user):
        for forum in forum.get_ancestors(ascending=True):
            # If no other topics are unread inside the considered forum, the latter should also
            # be marked as read.
            unread_topics = forum.topics.filter(
                Q(tracks__user=user, tracks__mark_time__lt=F('last_post_on')) |
                Q(forum__tracks__user=user, forum__tracks__mark_time__lt=F('last_post_on'), tracks__isnull=True))
            if unread_topics.exists():
                break

            # The topics that are marked as read inside the forum for the given user
            # wil be deleted while the forum track associated with the user must be
            # created or updated.
            TopicReadTrack.objects.filter(topic__forum=forum, user=user).delete()
            forum_track, _ = ForumReadTrack.objects.get_or_create(forum=forum, user=user)
            forum_track.save()



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django import template

# Local application / specific library imports
from machina.core.db.models import get_model
from machina.core.loading import get_class

TopicPollVote = get_model('forum_polls', 'TopicPollVote')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')
perm_handler = PermissionHandler()

get_anonymous_user_forum_key = get_class('forum_permission.shortcuts', 'get_anonymous_user_forum_key')

register = template.Library()


@register.filter
def has_been_completed_by(poll, user):
    """
    This will return a boolean indicating if the passed user has already
    voted in the given poll.
    Usage::
        {% if poll|has_been_completed_by:user %}...{% endif %}
    """
    user_votes = TopicPollVote.objects.filter(
        poll_option__poll=poll)
    if user.is_anonymous():
        forum_key = get_anonymous_user_forum_key(user)
        user_votes = user_votes.filter(anonymous_key=forum_key) if forum_key \
            else user_votes.none()
    else:
        user_votes = user_votes.filter(voter=user)
    return user_votes.exists()



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.contrib.auth import get_user_model
from django.db.models import Q

# Local application / specific library imports
from machina.conf import settings as machina_settings
from machina.core.db.models import get_model

ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class ForumPermissionChecker(object):
    """
    The ForumPermissionChecker allows to check forum permissions
    on Forum instances.
    """
    def __init__(self, user):
        self.user = user
        self._forum_perms_cache = {}

    def has_perm(self, perm, forum):
        """
        Checks if the considered user has given permission for the passed forum.
        """
        if not self.user.is_anonymous() and not self.user.is_active:
            return False
        elif self.user and self.user.is_superuser:
            return True
        return perm in self.get_perms(forum)

    def get_perms(self, forum):
        """
        Returns the list of permission codenames of all permissions for the given forum.
        """
        if not self.user.is_anonymous() and not self.user.is_active:
            return []
        user_model = get_user_model()
        user_groups_related_name = user_model.groups.field.related_query_name()

        if forum.id not in self._forum_perms_cache:
            if self.user and self.user.is_superuser:
                perms = list(ForumPermission.objects.values_list('codename', flat=True))
            elif self.user:
                default_auth_forum_perms = machina_settings.DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS

                user_kwargs_filter = {'anonymous_user': True} if self.user.is_anonymous() \
                    else {'user': self.user}

                user_perms = UserForumPermission.objects.select_related().filter(**user_kwargs_filter) \
                    .filter(Q(forum__isnull=True) | Q(forum=forum))

                globally_granted_user_perms = list(filter(lambda p: p.has_perm and p.forum is None, user_perms))
                globally_granted_user_perms = [p.permission.codename for p in globally_granted_user_perms]

                per_forum_granted_user_perms = list(filter(lambda p: p.has_perm and p.forum is not None, user_perms))
                per_forum_granted_user_perms = [p.permission.codename for p in per_forum_granted_user_perms]

                per_forum_nongranted_user_perms = list(filter(lambda p: not p.has_perm and p.forum is not None, user_perms))
                per_forum_nongranted_user_perms = [p.permission.codename for p in per_forum_nongranted_user_perms]

                if self.user.is_authenticated() and not globally_granted_user_perms:
                    globally_granted_user_perms = default_auth_forum_perms

                granted_user_perms = [c for c in globally_granted_user_perms if
                                      c not in per_forum_nongranted_user_perms] + per_forum_granted_user_perms
                granted_user_perms = set(granted_user_perms)

                perms = granted_user_perms

                if not self.user.is_anonymous():
                    group_perms = GroupForumPermission.objects.select_related() \
                        .filter(**{'group__{}'.format(user_groups_related_name): self.user}) \
                        .filter(Q(forum__isnull=True) | Q(forum=forum))

                    globally_granted_group_perms = list(filter(lambda p: p.has_perm and p.forum is None, group_perms))
                    globally_granted_group_perms = [p.permission.codename for p in globally_granted_group_perms]

                    per_forum_granted_group_perms = list(filter(lambda p: p.has_perm and p.forum is not None, group_perms))
                    per_forum_granted_group_perms = [p.permission.codename for p in per_forum_granted_group_perms]

                    per_forum_nongranted_group_perms = list(filter(lambda p: not p.has_perm and p.forum is not None, group_perms))
                    per_forum_nongranted_group_perms = [p.permission.codename for p in per_forum_nongranted_group_perms]

                    granted_group_perms = [c for c in globally_granted_group_perms if
                                           c not in per_forum_nongranted_group_perms] + per_forum_granted_group_perms
                    granted_group_perms = filter(lambda x: x not in per_forum_nongranted_user_perms, granted_group_perms)
                    granted_group_perms = set(granted_group_perms)

                    perms |= granted_group_perms

            self._forum_perms_cache[forum.id] = perms

        return self._forum_perms_cache[forum.id]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf.urls import include
from django.conf.urls import url
from django.utils.translation import ugettext_lazy as _

from machina.apps.forum_conversation.forum_attachments.app import application as attachments_app
from machina.apps.forum_conversation.forum_polls.app import application as polls_app
from machina.core.app import Application
from machina.core.loading import get_class


class BaseConversationApp(Application):
    name = 'forum_conversation'

    topic_view = get_class('forum_conversation.views', 'TopicView')
    topic_create_view = get_class('forum_conversation.views', 'TopicCreateView')
    topic_update_view = get_class('forum_conversation.views', 'TopicUpdateView')
    post_create_view = get_class('forum_conversation.views', 'PostCreateView')
    post_update_view = get_class('forum_conversation.views', 'PostUpdateView')
    post_delete_view = get_class('forum_conversation.views', 'PostDeleteView')

    def get_urls(self):
        urls = super(BaseConversationApp, self).get_urls()

        conversation_patterns = [
            url(_(r'^topic/(?P<slug>[\w-]+)-(?P<pk>\d+)/$'), self.topic_view.as_view(), name='topic'),

            url(_(r'^topic/create/$'), self.topic_create_view.as_view(), name='topic_create'),
            url(_(r'^topic/(?P<slug>[\w-]+)-(?P<pk>\d+)/update/$'), self.topic_update_view.as_view(), name='topic_update'),

            url(_(r'^topic/(?P<topic_slug>[\w-]+)-(?P<topic_pk>\d+)/post/create/$'),
                self.post_create_view.as_view(), name='post_create'),
            url(_(r'^topic/(?P<topic_slug>[\w-]+)-(?P<topic_pk>\d+)/(?P<pk>\d+)/post/update/$'),
                self.post_update_view.as_view(), name='post_update'),
            url(_(r'^topic/(?P<topic_slug>[\w-]+)-(?P<topic_pk>\d+)/(?P<pk>\d+)/post/delete/$'),
                self.post_delete_view.as_view(), name='post_delete'),
        ]

        urls += [
            url(_(r'forum/(?P<forum_slug>[\w-]+)-(?P<forum_pk>\d+)/'), include(conversation_patterns)),
        ]
        return urls


class PollsApp(Application):
    name = None
    polls_app = polls_app

    def get_urls(self):
        urls = super(PollsApp, self).get_urls()
        urls += [
            url(r'^', include(self.polls_app.urls)),
        ]
        return urls


class AttachmentsApp(Application):
    name = None
    attachments_app = attachments_app

    def get_urls(self):
        urls = super(AttachmentsApp, self).get_urls()
        urls += [
            url(r'^', include(self.attachments_app.urls)),
        ]
        return urls


class ConversationApp(BaseConversationApp, PollsApp, AttachmentsApp):
    """
    Composite class combining Conversation views with Polls views and Attachments views.
    """


application = ConversationApp()



# -*- coding: utf-8 -*-

# Standard library imports
# Third party imports
from django.contrib import admin

# Local application / specific library imports
from machina.core.db.models import get_model
from machina.models.fields import MarkupTextField
from machina.models.fields import MarkupTextFieldWidget

ForumProfile = get_model('forum_member', 'ForumProfile')


class ForumProfileAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'posts_count', )
    list_filter = ('posts_count', )
    list_display_links = ('id', 'user', )
    raw_id_fields = ('user', )
    search_fields = ('user__username',)

    formfield_overrides = {
        MarkupTextField: {'widget': MarkupTextFieldWidget},
    }


admin.site.register(ForumProfile, ForumProfileAdmin)



# -*- coding: utf-8 -*-

# Standard library imports
# Third party imports
from django.db.models import F
from django.dispatch import receiver

# Local application / specific library imports
from machina.apps.forum.signals import forum_viewed


@receiver(forum_viewed)
def update_forum_redirects_counter(sender, forum, user, request, response, **kwargs):
    """
    Receiver to handle the update of the link redirects counter associated with link forums.
    """
    if forum.is_link and forum.link_redirects:
        forum.link_redirects_count = F('link_redirects_count') + 1
        forum.save()



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.contrib import admin

# Local application / specific library imports
from machina.core.db.models import get_model

ForumReadTrack = get_model('forum_tracking', 'ForumReadTrack')
TopicReadTrack = get_model('forum_tracking', 'TopicReadTrack')


class ForumReadTrackAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'user', 'forum', 'mark_time',)
    list_filter = ('mark_time',)


class TopicReadTrackAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'user', 'topic', 'mark_time',)
    list_filter = ('mark_time',)


admin.site.register(ForumReadTrack, ForumReadTrackAdmin)
admin.site.register(TopicReadTrack, TopicReadTrackAdmin)



# -*- coding: utf-8 -*-

# Standard library imports
from collections import Iterable

# Third party imports
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import PermissionDenied
from django.contrib.auth.decorators import REDIRECT_FIELD_NAME
from django.http import HttpResponseRedirect
from django.utils.http import urlquote
from django.utils.six import string_types

# Local application / specific library imports


class PermissionRequiredMixin(object):
    """
    This view mixin verifies if the current user has the permissions specified by the
    'permission_required' attribute. This 'permissions check' behavior can be updated
    in the 'perform_permissions_check()' method.

    It provides the following workflow:

        The mixin try to see if the view and the current user passes the permission check.

        If the permission check fails and if the user isn't logged in, redirect to
        settings.LOGIN_URL passing the current absolute path in the qyery string. Example:
        /accounts/login/?next=/forum/3/

        If the permission check fails and if the user is logged in, return a 403 error
        page.

    The permissions will be tested against a specific instance provided either by a
    `get_object()` method or by an `object` attribute. In the case where the permissions
    should be checked against an instance that is not the one associated with a specific
    DetailView, it is possible to write a `get_controlled_object` method to which it will
    be given priority over the methods and attributes mentioned previously.
    """
    login_url = settings.LOGIN_URL
    permission_required = None
    redirect_field_name = REDIRECT_FIELD_NAME

    def get_required_permissions(self, request):
        """
        Returns the required permissions to access the considered object.
        """
        perms = []

        if not self.permission_required:
            return perms

        if isinstance(self.permission_required, string_types):
            perms = [self.permission_required, ]
        elif isinstance(self.permission_required, Iterable):
            perms = [perm for perm in self.permission_required]
        else:
            raise ImproperlyConfigured(
                '\'PermissionRequiredMixin\' requires \'permission_required\' '
                'attribute to be set to \'<app_label>.<permission codename>\' but is set to {} instead'.format(
                    self.permission_required)
            )
        return perms

    def perform_permissions_check(self, user, obj, perms):
        """
        Performs a permissions check in order to tell if the passed user
        can access the current view for the given object.
        By default, this method checks whether the given user has all the
        considered permissions in order to grant access. This behavior can
        be overridden in any subclass.
        """
        # Initializes a permission checker
        checker = self.request.forum_permission_handler._get_checker(user)
        # Check permissions
        return all(checker.has_perm(perm, obj) for perm in perms)

    def check_permissions(self, request):
        """
        Retrieve the controlled object and perform the permissions check.
        """
        obj = (hasattr(self, 'get_controlled_object') and self.get_controlled_object() or
               hasattr(self, 'get_object') and self.get_object() or getattr(self, 'object', None))
        user = request.user

        # Get the permissions to check
        perms = self.get_required_permissions(self)

        # Check permissions
        has_permissions = self.perform_permissions_check(user, obj, perms)

        if not has_permissions and not user.is_authenticated():
            return HttpResponseRedirect('{}?{}={}'.format(
                self.login_url,
                self.redirect_field_name,
                urlquote(request.get_full_path())
            ))
        elif not has_permissions:
            raise PermissionDenied

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs
        response = self.check_permissions(request)
        if response:
            return response
        return super(PermissionRequiredMixin, self).dispatch(request, *args, **kwargs)



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals
import os

# Third party imports
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

# Local application / specific library imports
from machina.conf import settings as machina_settings


@python_2_unicode_compatible
class AbstractAttachment(models.Model):
    """
    Represents a post attachment. An attachment is always linked to a post.
    """
    post = models.ForeignKey('forum_conversation.Post', verbose_name=_('Post'), related_name='attachments')
    file = models.FileField(verbose_name=_('File'), upload_to=machina_settings.ATTACHMENT_FILE_UPLOAD_TO)
    comment = models.CharField(max_length=255, verbose_name=_('Comment'), blank=True, null=True)

    class Meta:
        abstract = True
        app_label = 'forum_attachments'
        verbose_name = _('Attachment')
        verbose_name_plural = _('Attachments')

    def __str__(self):
        return '{}'.format(self.post.subject)

    @property
    def filename(self):
        return os.path.basename(self.file.name)

    def get_absolute_url(self):
        from django.core.urlresolvers import reverse
        return reverse('forum_conversation:attachment', kwargs={'pk': str(self.id)})



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.utils.translation import ugettext_lazy as _

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class

Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')
TopicPoll = get_model('forum_polls', 'TopicPoll')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')

get_anonymous_user_forum_key = get_class('forum_permission.shortcuts', 'get_anonymous_user_forum_key')


class PostForm(forms.ModelForm):
    class Meta:
        model = Post
        fields = ['subject', 'content', 'username', 'update_reason', ]

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        self.user_ip = kwargs.pop('user_ip', None)
        self.forum = kwargs.pop('forum', None)
        self.topic = kwargs.pop('topic', None)

        self.perm_handler = PermissionHandler()

        super(PostForm, self).__init__(*args, **kwargs)

        # Updates the 'subject' and 'content' fields attributes
        self.fields['subject'].widget.attrs['placeholder'] = _('Enter your subject')
        self.fields['content'].label = _('Message')
        self.fields['content'].widget.attrs['placeholder'] = _('Enter your message')

        # Handles anonymous users
        if self.user and self.user.is_anonymous():
            self.fields['username'].required = True
        else:
            # The 'username' field is not really usefull if the user is
            # authenticated
            del self.fields['username']

        # Handles the definition of a default subject if we are
        # considering an answer
        if not self.instance.pk and self.topic:
            self.fields['subject'].initial = '{} {}'.format(
                machina_settings.TOPIC_ANSWER_SUBJECT_PREFIX,
                self.topic.subject)

        # Delete the 'update_reason' field if we are
        # considering a post update
        if not self.instance.pk:
            del self.fields['update_reason']

    def clean(self):
        if not self.user.is_anonymous():
            self.instance.poster = self.user
        else:
            self.instance.anonymous_key = get_anonymous_user_forum_key(self.user)
        return super(PostForm, self).clean()

    def save(self, commit=True):
        if self.instance.pk:
            # First handle updates
            post = super(PostForm, self).save(commit=False)
            post.updated_by = self.user
            post.updates_count = F('updates_count') + 1
        else:
            post = Post(
                topic=self.topic,
                poster_ip=self.user_ip,
                subject=self.cleaned_data['subject'],
                approved=self.perm_handler.can_post_without_approval(self.forum, self.user),
                content=self.cleaned_data['content'])
            if not self.user.is_anonymous():
                post.poster = self.user
            else:
                post.username = self.cleaned_data['username']
                post.anonymous_key = get_anonymous_user_forum_key(self.user)

        if commit:
            post.save()

        return post


class TopicForm(PostForm):
    topic_type = forms.ChoiceField(label=_('Post topic as'), choices=Topic.TYPE_CHOICES, required=False)

    def __init__(self, *args, **kwargs):
        super(TopicForm, self).__init__(*args, **kwargs)

        # Perform some checks before doing anything
        self.can_add_stickies = self.perm_handler.can_add_stickies(self.forum, self.user)
        self.can_add_announcements = self.perm_handler.can_add_announcements(self.forum, self.user)
        self.can_create_polls = self.perm_handler.can_create_polls(self.forum, self.user)

        if not self.can_add_stickies:
            choices = filter(
                lambda t: t[0] != Topic.TYPE_CHOICES.topic_sticky,
                self.fields['topic_type'].choices)
            self.fields['topic_type'].choices = choices
        if not self.can_add_announcements:
            choices = filter(
                lambda t: t[0] != Topic.TYPE_CHOICES.topic_announce,
                self.fields['topic_type'].choices)
            self.fields['topic_type'].choices = choices

        # Append polls fields to the form if the user is allowed to create such things
        if self.can_create_polls:
            self.fields['poll_question'] = forms.CharField(
                label=_('Poll question'), required=False,
                help_text=_('Enter a question to associate a poll with the topic or leave blank to not create a poll.'),
                max_length=TopicPoll._meta.get_field('question').max_length)
            self.fields['poll_max_options'] = forms.IntegerField(
                label=_('Maximum number of poll options per user'), required=False,
                help_text=_('This is the number of options each user may select when voting.'),
                validators=TopicPoll._meta.get_field('max_options').validators,
                initial=1)
            self.fields['poll_duration'] = forms.IntegerField(
                label=_('For how many days the poll should be run?'), required=False,
                help_text=_('Enter 0 or leave blank for a never ending poll.'),
                min_value=0, initial=0)
            self.fields['poll_user_changes'] = forms.BooleanField(
                label=_('Allow re-voting?'), required=False,
                help_text=_('If enabled users are able to change their vote.'),
                initial=False)

        # Set the initial values
        try:
            if hasattr(self.instance, 'topic'):
                self.fields['topic_type'].initial = self.instance.topic.type

                if self.can_create_polls and self.instance.topic.poll is not None:
                    self.fields['poll_question'].initial = self.instance.topic.poll.question
                    self.fields['poll_max_options'].initial = self.instance.topic.poll.max_options
                    self.fields['poll_duration'].initial = self.instance.topic.poll.duration
                    self.fields['poll_user_changes'].initial = self.instance.topic.poll.user_changes
        except ObjectDoesNotExist:
            pass

    def save(self, commit=True):
        if not self.instance.pk:
            # First, handle topic creation
            if 'topic_type' in self.cleaned_data and len(self.cleaned_data['topic_type']):
                topic_type = self.cleaned_data['topic_type']
            else:
                topic_type = Topic.TYPE_CHOICES.topic_post

            topic = Topic(
                forum=self.forum,
                subject=self.cleaned_data['subject'],  # The topic's name is the post's name
                type=topic_type,
                status=Topic.STATUS_CHOICES.topic_unlocked,
                approved=self.perm_handler.can_post_without_approval(self.forum, self.user))
            if not self.user.is_anonymous():
                topic.poster = self.user
            self.topic = topic
            if commit:
                topic.save()
        else:
            if 'topic_type' in self.cleaned_data and len(self.cleaned_data['topic_type']):
                if self.instance.topic.type != self.cleaned_data['topic_type']:
                    self.instance.topic.type = self.cleaned_data['topic_type']
                    self.instance.topic._simple_save()

        return super(TopicForm, self).save(commit)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _


# PILImage
try:
    # Try from the Pillow (or one variant of PIL) install location first.
    from PIL import Image as PILImage
except ImportError as err:
    try:
        # If that failed, try the alternate import syntax for PIL.
        import Image as PILImage  # noqa
    except ImportError as err:
        # Neither worked, so it's likely not installed.
        raise ImproperlyConfigured(
            _('Neither Pillow nor PIL could be imported: {}').format(err)
        )



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals
from collections import OrderedDict

# Third party imports
# Local application / specific library imports


class PermissionConfig(object):
    """
    This class defines the default configuration of forum permissions. It is used
    to create the related ForumPermission instances on a syncdb or migrate operation.
    Moreover the permission scopes are used in admin pages to display permission forms
    in a readable way.
    """
    permissions = [
        # Forums
        {
            'fields': {
                'codename': 'can_see_forum',
                'name': 'Can see forum',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'forum',
        },
        {
            'fields': {
                'codename': 'can_read_forum',
                'name': 'Can read forum',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'forum',
        },
        # Topics & posts
        {
            'fields': {
                'codename': 'can_start_new_topics',
                'name': 'Can start new topics',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_reply_to_topics',
                'name': 'Can reply to topics',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_post_announcements',
                'name': 'Can post announcements',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_post_stickies',
                'name': 'Can post stickies',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_delete_own_posts',
                'name': 'Can delete own posts',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_edit_own_posts',
                'name': 'Can edit own posts',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        {
            'fields': {
                'codename': 'can_post_without_approval',
                'name': 'Can post without approval',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'conversation',
        },
        # Polls
        {
            'fields': {
                'codename': 'can_create_polls',
                'name': 'Can create polls',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'polls',
        },
        {
            'fields': {
                'codename': 'can_vote_in_polls',
                'name': 'Can vote in polls',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'polls',
        },
        # Attachments
        {
            'fields': {
                'codename': 'can_attach_file',
                'name': 'Can attach file',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'attachments',
        },
        {
            'fields': {
                'codename': 'can_download_file',
                'name': 'Can download file',
                'is_local': True,
                'is_global': True,
            },
            'scope': 'attachments',
        },
        # Moderation
        {
            'fields': {
                'codename': 'can_lock_topics',
                'name': 'Can lock topics',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
        {
            'fields': {
                'codename': 'can_move_topics',
                'name': 'Can move topics',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
        {
            'fields': {
                'codename': 'can_edit_posts',
                'name': 'Can edit posts',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
        {
            'fields': {
                'codename': 'can_delete_posts',
                'name': 'Can delete posts',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
        {
            'fields': {
                'codename': 'can_approve_posts',
                'name': 'Can approve posts',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
        {
            'fields': {
                'codename': 'can_reply_to_locked_topics',
                'name': 'Can add posts in locked topics',
                'is_local': True,
                'is_global': False,
            },
            'scope': 'moderation',
        },
    ]

    scopes = list(OrderedDict.fromkeys([p['scope'] for p in permissions]))



# -*- coding: utf-8 -*-

# Standard library imports
# Third party imports
from django.core import validators

# Local application / specific library imports
from machina.conf import settings as machina_settings


poll_max_options = [
    validators.MinValueValidator(1),
    validators.MaxValueValidator(machina_settings.POLL_MAX_OPTIONS_PER_USER),
]



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django import forms
from django.utils.translation import ugettext_lazy as _

# Local application / specific library imports
from machina.core.db.models import get_model
from machina.core.loading import get_class
from machina.forms.widgets import SelectWithDisabled

Forum = get_model('forum', 'Forum')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class TopicMoveForm(forms.Form):
    forum = forms.ChoiceField(
        label=_('Select a destination forum'),
        widget=SelectWithDisabled)
    lock_topic = forms.BooleanField(
        label=_('Lock topic'), required=False)

    def __init__(self, *args, **kwargs):
        self.topic = kwargs.pop('topic', None)
        self.user = kwargs.pop('user', None)
        self.perm_handler = PermissionHandler()

        super(TopicMoveForm, self).__init__(*args, **kwargs)

        self.allowed_forums = self.perm_handler.get_target_forums_for_moved_topics(self.user)
        forum_choices = []

        for f in self.allowed_forums:
            if f.is_category or f.id == self.topic.forum.id:
                forum_choices.append((
                    f.id,
                    {
                        'label': '{} {}'.format('-' * f.margin_level, f.name),
                        'disabled': True,
                    }))
            else:
                forum_choices.append((f.id, '{} {}'.format('-' * f.margin_level, f.name)))

        if self.topic.is_locked:
            self.fields['lock_topic'].initial = True

        self.fields['forum'].choices = forum_choices

    def clean_forum(self):
        forum_id = self.cleaned_data['forum']

        if forum_id:
            forum = Forum.objects.get(pk=forum_id)
            if forum.is_category or forum.is_link or forum.id == self.topic.forum.id:
                raise forms.ValidationError('You cannot select this forum as a destination')

        return forum



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.dispatch import receiver

# Local application / specific library imports
from machina.core.loading import get_class

topic_viewed = get_class('forum_conversation.signals', 'topic_viewed')


@receiver(topic_viewed)
def update_user_trackers(sender, topic, user, request, response, **kwargs):
    """
    Receiver to mark a topic being viewed as read. This can result in marking
    the related forum tracker as read.
    """
    TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')  # noqa
    track_handler = TrackingHandler()
    track_handler.mark_topic_read(topic, user)



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext_lazy as _
from django.views.generic import DetailView
from django.views.generic import ListView
from django.views.generic import UpdateView

# Local application / specific library imports
from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class

Forum = get_model('forum', 'Forum')
ForumProfile = get_model('forum_member', 'ForumProfile')
Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')

ForumProfileForm = get_class('forum_member.forms', 'ForumProfileForm')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')
perm_handler = PermissionHandler()


class UserTopicsView(ListView):
    """
    Provides a list of all the topics in which the current user has
    posted messages.
    """
    template_name = 'forum_member/user_topics_list.html'
    context_object_name = 'topics'
    paginate_by = machina_settings.FORUM_TOPICS_NUMBER_PER_PAGE

    def get_queryset(self):
        forums = perm_handler.forum_list_filter(
            Forum.objects.all(), self.request.user)
        topics = Topic.objects.filter(
            forum__in=forums,
            posts__poster_id=self.request.user.id,
            posts__approved=True)
        return topics.order_by('-updated')

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(UserTopicsView, self).dispatch(request, *args, **kwargs)


class ForumProfileDetailView(DetailView):
    """
    Shows a user's forum profile.
    """
    template_name = 'forum_member/forum_profile_detail.html'
    context_object_name = 'profile'

    def get_queryset(self):
        user_model = get_user_model()
        return user_model.objects.all()

    def get_object(self, queryset=None):
        user = super(ForumProfileDetailView, self).get_object(queryset)
        profile, dummy = ForumProfile.objects.get_or_create(user=user)
        return profile

    def get_context_data(self, **kwargs):
        context = super(ForumProfileDetailView, self).get_context_data(**kwargs)

        # Computes the number of topics added by the considered member
        context['topics_count'] = Topic.objects.filter(
            approved=True, poster=self.object.user).count()

        # Fetches the recent posts added by the considered user
        forums = self.request.forum_permission_handler.forum_list_filter(
            Forum.objects.all(), self.request.user)
        recent_posts = Post.approved_objects.filter(
            topic__forum__in=forums, poster=self.object.user).order_by('-created')
        context['recent_posts'] = recent_posts[:machina_settings.PROFILE_RECENT_POSTS_NUMBER]

        return context


class ForumProfileUpdateView(UpdateView):
    """
    Allows the current user to update its forum profile.
    """
    template_name = 'forum_member/forum_profile_update.html'
    form_class = ForumProfileForm
    success_message = _('The profile has been edited successfully.')

    def get_object(self, queryset=None):
        profile, dummy = ForumProfile.objects.get_or_create(user=self.request.user)
        return profile

    def form_valid(self, form):
        response = super(ForumProfileUpdateView, self).form_valid(form)
        messages.success(self.request, self.success_message)
        return response

    def get_success_url(self):
        return reverse('forum_member:profile_update')

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(ForumProfileUpdateView, self).dispatch(request, *args, **kwargs)



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import unicode_literals

# Third party imports
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils.encoding import force_text
from django.utils.encoding import python_2_unicode_compatible
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _
from model_utils import Choices

# Local application / specific library imports
from machina.core.loading import get_class
from machina.core.shortcuts import refresh
from machina.models.abstract_models import DatedModel
from machina.models.fields import MarkupTextField

ApprovedManager = get_class('forum_conversation.managers', 'ApprovedManager')


TOPIC_TYPES = Choices(
    (0, 'topic_post', _('Default topic')),
    (1, 'topic_sticky', _('Sticky')),
    (2, 'topic_announce', _('Announce')),
)

TOPIC_STATUSES = Choices(
    (0, 'topic_unlocked', _('Topic unlocked')),
    (1, 'topic_locked', _('Topic locked')),
    (2, 'topic_moved', _('Topic moved')),
)


@python_2_unicode_compatible
class AbstractTopic(DatedModel):
    """
    Represents a forum topic.
    """
    forum = models.ForeignKey('forum.Forum', verbose_name=_('Topic forum'), related_name='topics')
    poster = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_('Poster'), blank=True, null=True)

    # The subject of the thread should correspond to the one associated with the first post
    subject = models.CharField(max_length=255, verbose_name=_('Subject'))
    slug = models.SlugField(max_length=300, verbose_name=_('Slug'))

    # Sticky, Announce, Global topic or Default topic ; that's what a topic can be
    TYPE_CHOICES = TOPIC_TYPES
    type = models.PositiveSmallIntegerField(choices=TOPIC_TYPES, verbose_name=_('Topic type'), db_index=True)

    # A topic can be locked, unlocked or moved
    STATUS_CHOICES = TOPIC_STATUSES
    status = models.PositiveIntegerField(choices=TOPIC_STATUSES, verbose_name=_('Topic status'), db_index=True)

    # A topic can be approved before publishing ; defaults to True. The value of this flag
    # should correspond to the one associated with the first post
    approved = models.BooleanField(verbose_name=_('Approved'), default=True)

    # The number of posts included in this topic (only those that are approved)
    posts_count = models.PositiveIntegerField(verbose_name=_('Posts count'), editable=False, blank=True, default=0)

    # The number of time the topic has been viewed
    views_count = models.PositiveIntegerField(verbose_name=_('Views count'), editable=False, blank=True, default=0)

    # The date of the latest post
    last_post_on = models.DateTimeField(verbose_name=_('Last post added on'), blank=True, null=True)

    # Many users can subscribe to this topic
    subscribers = models.ManyToManyField(settings.AUTH_USER_MODEL, related_name='subscriptions', verbose_name=_('Subscribers'), blank=True)

    objects = models.Manager()
    approved_objects = ApprovedManager()

    class Meta:
        abstract = True
        app_label = 'forum_conversation'
        ordering = ['-type', '-last_post_on', ]
        get_latest_by = 'last_post_on'
        verbose_name = _('Topic')
        verbose_name_plural = _('Topics')

    def __str__(self):
        if self.posts.exists():
            return self.posts.all().order_by('created')[0].subject
        else:
            return str(self.id)

    @property
    def is_topic(self):
        """
        Returns True if the topic is a default topic.
        """
        return self.type == self.TYPE_CHOICES.topic_post

    @property
    def is_sticky(self):
        """
        Returns True if the topic is a sticky topic.
        """
        return self.type == self.TYPE_CHOICES.topic_sticky

    @property
    def is_announce(self):
        """
        Returns True if the topic is an announce.
        """
        return self.type == self.TYPE_CHOICES.topic_announce

    @property
    def is_locked(self):
        """
        Returns True if the topic is locked.
        """
        return self.status == self.STATUS_CHOICES.topic_locked

    @property
    def first_post(self):
        """
        Try to fetch the first post associated with the current topic and caches it to
        lighten the next request.
        """
        if not hasattr(self, '_first_post'):
            posts = self.posts.select_related('poster').all().order_by('created')
            self._first_post = posts[0] if len(posts) else None
        return self._first_post

    @property
    def last_post(self):
        """
        Try to fetch the last post associated with the current topic and caches it to
        lighten the next request.
        """
        if not hasattr(self, '_last_post'):
            posts = self.posts.select_related('poster').filter(approved=True).order_by('-created')
            self._last_post = posts[0] if len(posts) else None
        return self._last_post

    def clean(self):
        super(AbstractTopic, self).clean()
        if self.forum.is_category or self.forum.is_link:
            raise ValidationError(_('A topic can not be associated with a category or a link forum'))

    def save(self, *args, **kwargs):
        # It is vital to track the changes of the forum associated with a topic in order to
        # maintain counters up-to-date.
        old_instance = None
        if self.pk:
            old_instance = self.__class__._default_manager.get(pk=self.pk)

        # Update the slug field
        self.slug = slugify(force_text(self.subject))

        # Do the save
        super(AbstractTopic, self).save(*args, **kwargs)

        # If any change has been made to the parent forum, trigger the update of the counters
        if old_instance and old_instance.forum != self.forum:
            self.update_trackers()
            # The previous parent forum counters should also be updated
            if old_instance.forum:
                old_forum = refresh(old_instance.forum)
                old_forum.update_trackers()

    def _simple_save(self, *args, **kwargs):
        """
        Calls the parent save method in order to avoid the checks for topic forum changes
        which can result in triggering a new update of the counters associated with the
        current topic.
        This allow the database to not be hit by such checks during very common and regular
        operations such as those provided by the update_trackers function; indeed these operations
        will never result in an update of a topic's forum.
        """
        super(AbstractTopic, self).save(*args, **kwargs)

    def delete(self, using=None):
        super(AbstractTopic, self).delete(using)
        self.forum.update_trackers()

    def update_trackers(self):
        """
        Updates the posts count, the update date and the link toward the last post
        associated with the current topic.
        """
        self.posts_count = self.posts.filter(approved=True).count()
        posts = self.posts.filter(approved=True).order_by('-created')
        self._last_post = posts[0] if posts.exists() else None
        self.last_post_on = self._last_post.created if self._last_post else None
        self._simple_save()
        # Trigger the forum-level trackers update
        self.forum.update_trackers()

    def get_absolute_url(self):
        from django.core.urlresolvers import reverse
        return reverse('forum_conversation:topic', kwargs={
            'forum_slug': self.forum.slug, 'forum_pk': self.forum.pk,
            'slug': self.slug, 'pk': str(self.id)})


@python_2_unicode_compatible
class AbstractPost(DatedModel):
    """
    Represents a forum post. A forum post is always linked to a topic.
    """
    topic = models.ForeignKey('forum_conversation.Topic', verbose_name=_('Topic'), related_name='posts')
    poster = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='posts', verbose_name=_('Poster'), blank=True, null=True)
    poster_ip = models.GenericIPAddressField(verbose_name=_('Poster IP address'), blank=True, null=True, default='2002::0')
    anonymous_key = models.CharField(max_length=100, verbose_name=_('Anonymous user forum key'), blank=True, null=True)

    # Each post can have its own subject. The subject of the thread corresponds to the
    # one associated with the first post
    subject = models.CharField(verbose_name=_('Subject'), max_length=255)

    # Content
    content = MarkupTextField(verbose_name=_('Content'))

    # Username: if the user creating a topic post is not authenticated, he must enter a username
    username = models.CharField(verbose_name=_('Username'), max_length=155, blank=True, null=True)

    # A post can be approved before publishing ; defaults to True
    approved = models.BooleanField(verbose_name=_('Approved'), default=True)

    # A post can be edited for several reason (eg. moderation) ; the reason why it has been updated can be specified
    update_reason = models.CharField(max_length=255, verbose_name=_('Update reason'), blank=True, null=True)

    # Tracking data
    updated_by = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_('Lastly updated by'), editable=False, blank=True, null=True)
    updates_count = models.PositiveIntegerField(verbose_name=_('Updates count'), editable=False, blank=True, default=0)

    objects = models.Manager()
    approved_objects = ApprovedManager()

    class Meta:
        abstract = True
        app_label = 'forum_conversation'
        ordering = ['created', ]
        get_latest_by = 'created'
        verbose_name = _('Post')
        verbose_name_plural = _('Posts')

    def __str__(self):
        return self.subject

    @property
    def is_topic_head(self):
        """
        Returns True if the post is the first post of the topic.
        """
        return self.topic.first_post.id == self.id

    @property
    def is_topic_tail(self):
        """
        Returns True if the post is the last post of the topic.
        """
        return self.topic.last_post.id == self.id

    @property
    def position(self):
        """
        Returns an integer corresponding to the position of the post in the topic.
        """
        position = self.topic.posts.filter(Q(created__lt=self.created) | Q(id=self.id)).count()
        return position

    def clean(self):
        super(AbstractPost, self).clean()

        # At least a poster (user) or a session key must be associated with
        # the post.
        if self.poster is None and self.anonymous_key is None:
            raise ValidationError(_('A user id or an anonymous key must be associated with a post.'))
        if self.poster and self.anonymous_key:
            raise ValidationError(_('A user id or an anonymous key must be associated with a post, '
                                    'but not both.'))

        if self.anonymous_key and not self.username:
            raise ValidationError(_('A username must be specified if the poster is anonymous'))

    def save(self, *args, **kwargs):
        super(AbstractPost, self).save(*args, **kwargs)

        # Ensures that the subject of the thread corresponds to the one associated
        # with the first post. Do the same with the 'approved' flag.
        if self.is_topic_head:
            if self.subject != self.topic.subject or self.approved != self.topic.approved:
                self.topic.subject = self.subject
                self.topic.approved = self.approved

        # Trigger the topic-level trackers update
        self.topic.update_trackers()

    def delete(self, using=None):
        if self.is_topic_head and self.is_topic_tail:
            # The default way of operating is to trigger the deletion of the associated topic
            # only if the considered post is the only post embedded in the topic
            self.topic.delete()
        else:
            super(AbstractPost, self).delete(using)
            self.topic.update_trackers()



# -*- coding: utf-8 -*-

# Standard library imports
from __future__ import division
from __future__ import unicode_literals

# Third party imports
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

# Local application / specific library imports
from machina.apps.forum_conversation.forum_polls import validators
from machina.models.abstract_models import DatedModel


@python_2_unicode_compatible
class AbstractTopicPoll(DatedModel):
    """
    Represents a poll embedded in a forum topic.
    """
    topic = models.OneToOneField('forum_conversation.Topic', verbose_name=_('Topic'), related_name='poll')

    # A poll is defined by a single question
    question = models.CharField(max_length=255, verbose_name=_('Poll question'))

    # A poll can have a duration
    duration = models.PositiveIntegerField(verbose_name=_('Poll duration, in days'), blank=True, null=True)

    # Users can possibly select more than one option associated with a poll
    max_options = models.PositiveSmallIntegerField(
        verbose_name=_('Maximum number of poll options per user'),
        validators=validators.poll_max_options, default=1)

    # Are users allowed to change their votes ?
    user_changes = models.BooleanField(verbose_name=_('Allow vote changes'), default=False)

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        ordering = ['-updated', ]
        get_latest_by = 'updated'
        verbose_name = _('Topic poll')
        verbose_name_plural = _('Topic polls')

    def __str__(self):
        return '{}'.format(self.topic.subject)

    @cached_property
    def votes(self):
        """
        Returns all the votes related to this topic poll.
        """
        votes = []
        for option in self.options.all():
            votes += option.votes.all()
        return votes


@python_2_unicode_compatible
class AbstractTopicPollOption(models.Model):
    """
    Represents a poll option.
    """
    poll = models.ForeignKey('forum_polls.TopicPoll', verbose_name=_('Poll'), related_name='options')
    text = models.CharField(max_length=255, verbose_name=_('Poll option text'))

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        verbose_name = _('Topic poll option')
        verbose_name_plural = _('Topic poll options')

    def __str__(self):
        return '{} - {}'.format(self.poll, self.text)

    @property
    def percentage(self):
        if not hasattr(self, '_percentage'):
            self._percentage = (self.votes.count() / (len(self.poll.votes) or 1)) * 100
        return self._percentage


@python_2_unicode_compatible
class AbstractTopicPollVote(models.Model):
    """
    Represents a poll vote.
    """
    voter = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_('Voter'), related_name='poll_votes', blank=True, null=True)
    anonymous_key = models.CharField(
        max_length=100, verbose_name=_('Anonymous user forum key'), blank=True, null=True)

    poll_option = models.ForeignKey('forum_polls.TopicPollOption', verbose_name=_('Poll option'), related_name='votes')
    timestamp = models.DateTimeField(auto_now_add=True, verbose_name=_('Vote\'s date'))

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        verbose_name = _('Topic poll vote')
        verbose_name_plural = _('Topic poll votes')

    def __str__(self):
        return '{} - {}'.format(self.poll_option, self.voter)

    def clean(self):
        super(AbstractTopicPollVote, self).clean()

        # At least a poster (user) or a session key must be associated with
        # the vote instance.
        if self.voter is None and self.anonymous_key is None:
            raise ValidationError(_('A user id or an anonymous key must be used.'))
        if self.voter and self.anonymous_key:
            raise ValidationError(_('A user id or an anonymous key must be used, but not both.'))
